import functools
import inspect
import json
import os
import os.path as osp
import pickle
import uuid
from typing import Any, List, Literal
from loguru import logger
import cachetools
import xxhash

from .utils_io import dump_json_or_pickle, load_json_or_pickle
from .utils_misc import mkdir_or_exist

SPEED_CACHE_DIR = osp.join(osp.expanduser("~"), ".cache/av")
LRU_MEM_CACHE = cachetools.LRUCache(maxsize=128_000)
from threading import Lock

thread_locker = Lock()

# Add two locks for thread-safe cache access
disk_lock = Lock()
mem_lock = Lock()


def fast_serialize(x: Any) -> bytes:
    try:
        return json.dumps(x, sort_keys=True).encode("utf-8")
    except (TypeError, ValueError):
        return pickle.dumps(x, protocol=pickle.HIGHEST_PROTOCOL)


def identify(x: Any) -> str:
    return xxhash.xxh64_hexdigest(fast_serialize(x), seed=0)


def identify_uuid(x: Any) -> str:
    data = fast_serialize(x)
    hash_obj = xxhash.xxh128(data, seed=0)
    return str(uuid.UUID(bytes=hash_obj.digest()))


def _get_source(func):
    code = inspect.getsource(func)
    for r in [" ", "\n", "\t", "\r"]:
        code = code.replace(r, "")
    return code


def _compute_func_id(func, args, kwargs, ignore_self, cache_key, keys):
    func_source = _get_source(func)
    if keys:
        arg_spec = inspect.getfullargspec(func).args
        used_args = {arg_spec[i]: arg for i, arg in enumerate(args)}
        used_args.update(kwargs)
        values = [used_args[k] for k in keys if k in used_args]
        if not values:
            return None, None, None
        dir_path = f"{func.__name__}_{identify(func_source)}"
        key_id = f"{'_'.join(keys)}_{identify(values)}.pkl"
        return func_source, dir_path, key_id

    if cache_key and cache_key in kwargs:
        fid = [func_source, kwargs[cache_key]]
    elif (
        inspect.getfullargspec(func).args
        and inspect.getfullargspec(func).args[0] == "self"
        and ignore_self
    ):
        fid = (func_source, args[1:], kwargs)
    else:
        fid = (func_source, args, kwargs)
    return func_source, "funcs", f"{identify(fid)}.pkl"


def _disk_memoize(func, keys, cache_dir, ignore_self, verbose, cache_key):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        # Compute cache path as before
        func_source, sub_dir, key_id = _compute_func_id(func, args, kwargs, ignore_self, cache_key, keys)
        if func_source is None:
            return func(*args, **kwargs)
        if sub_dir == "funcs":
            cache_path = osp.join(cache_dir, sub_dir, func.__name__, key_id)
        else:
            cache_path = osp.join(cache_dir, sub_dir, key_id)
        mkdir_or_exist(osp.dirname(cache_path))
        
        # First check with disk lock
        with disk_lock:
            if osp.exists(cache_path):
                logger.debug(f"Cache HIT for {func.__name__}, key={cache_path}")
                return load_json_or_pickle(cache_path)
        
        result = func(*args, **kwargs)
        
        # Write result under disk lock to avoid race conditions
        with disk_lock:
            if not osp.exists(cache_path):
                logger.debug(f"Cache MISS for {func.__name__}, key={cache_path}")
                dump_json_or_pickle(result, cache_path)
        return result
    return wrapper


def _memory_memoize(func, size, keys, ignore_self, cache_key):
    global LRU_MEM_CACHE
    if LRU_MEM_CACHE.maxsize != size:
        LRU_MEM_CACHE = cachetools.LRUCache(maxsize=size)
    
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func_source, sub_dir, key_id = _compute_func_id(func, args, kwargs, ignore_self, cache_key, keys)
        if func_source is None:
            return func(*args, **kwargs)
        name = identify((func_source, sub_dir, key_id))
        
        if not hasattr(func, "_mem_cache"):
            func._mem_cache = LRU_MEM_CACHE
        
        with mem_lock:
            if name in func._mem_cache:
                logger.debug(f"Cache HIT (memory) for {func.__name__}, key={name}")
                return func._mem_cache[name]
        
        result = func(*args, **kwargs)
        
        with mem_lock:
            if name not in func._mem_cache:
                func._mem_cache[name] = result
        return result
    return wrapper


def _both_memoize(func, keys, cache_dir, ignore_self, verbose, cache_key):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        func_source, sub_dir, key_id = _compute_func_id(func, args, kwargs, ignore_self, cache_key, keys)
        if func_source is None:
            return func(*args, **kwargs)
        
        mem_key = identify((func_source, sub_dir, key_id))
        if not hasattr(func, "_mem_cache"):
            func._mem_cache = LRU_MEM_CACHE

        with mem_lock:
            if mem_key in func._mem_cache:
                logger.debug(f"Cache HIT (memory) for {func.__name__}, key={mem_key}")
                return func._mem_cache[mem_key]

        if sub_dir == "funcs":
            cache_path = osp.join(cache_dir, sub_dir, func.__name__, key_id)
        else:
            cache_path = osp.join(cache_dir, sub_dir, key_id)
        mkdir_or_exist(osp.dirname(cache_path))
        
        with disk_lock:
            if osp.exists(cache_path):
                logger.debug(f"Cache HIT (disk) for {func.__name__}, key={cache_path}")
                result = load_json_or_pickle(cache_path)
                with mem_lock:
                    func._mem_cache[mem_key] = result
                return result

        result = func(*args, **kwargs)
        
        with disk_lock:
            if not osp.exists(cache_path):
                logger.debug(f"Cache MISS for {func.__name__}, key={cache_path}")
                dump_json_or_pickle(result, cache_path)
        with mem_lock:
            func._mem_cache[mem_key] = result
        return result
    return wrapper


def memoize(
    _func=None,
    *,
    keys=None,
    cache_dir=SPEED_CACHE_DIR,
    cache_type: Literal["memory", "disk", "both"] = "both",
    size=128_000,
    ignore_self=True,
    verbose=False,
    cache_key=None,
):
    logger.info(f"cache_dir: {cache_dir}, cache_type: {cache_type}")

    def decorator(func):
        if cache_type == "memory":
            return _memory_memoize(func, size, keys, ignore_self, cache_key)
        elif cache_type == "disk":
            return _disk_memoize(func, keys, cache_dir, ignore_self, verbose, cache_key)
        return _both_memoize(func, keys, cache_dir, ignore_self, verbose, cache_key)

    if _func is None:
        return decorator
    return decorator(_func)


__all__ = ["memoize", "identify", "identify_uuid"]



import inspect
import os
import threading
import time
from collections import defaultdict
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Callable, List, Tuple
from loguru import logger
from speedy_utils.common.utils_print import fprint
from speedy_utils.multi_worker._handle_inputs import handle_inputs
from tqdm import tqdm
import signal
import functools


class TimeoutError(Exception):
    """Custom exception to be raised when a function times out."""

    pass


def timeout(t=30, error_message="Function call timed out"):
    """
    Decorator that raises a TimeoutError if the decorated function does not finish within 't' seconds.

    Parameters:
    - t: Timeout duration in seconds (default 30)
    - error_message: Message to include in the TimeoutError (default "Function call timed out")
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            # Define the handler for the SIGALRM signal.
            def _handle_timeout(signum, frame):
                raise TimeoutError(error_message)

            # Set the SIGALRM signal handler to our timeout handler.
            signal.signal(signal.SIGALRM, _handle_timeout)
            # Schedule the SIGALRM signal to be sent after t seconds.
            signal.alarm(t)
            try:
                result = func(*args, **kwargs)
            finally:
                # Cancel the alarm so it doesn't trigger if the function returned in time.
                signal.alarm(0)
            return result

        return wrapper

    return decorator


class RunErr(Exception):
    def __init__(self, error_type: str, message: str) -> None:
        self.error_type = error_type
        self.message = message

    def to_dict(self) -> dict:
        return {"error_type": self.error_type, "message": self.message}

    def __str__(self) -> str:
        formated = f"{self.error_type}: {self.message}"
        return formated


def _process_single_task(
    func: Callable, idx: int, inp: Any, stop_event: threading.Event
) -> Tuple[int, Any]:
    """Process a single task and handle exceptions."""
    if stop_event.is_set():
        return idx, None

    try:
        ret = func(inp) if not isinstance(inp, dict) else func(**inp)
        return idx, ret
    except Exception as e:
        import traceback

        return idx, RunErr(type(e).__name__, traceback.format_exc())


def _handle_results(
    future: Any,
    results: List[Any],
    stop_event: threading.Event,
    stop_on_error: bool,
    result_counter: defaultdict,
    pbar: tqdm,
) -> None:
    """Handle the results from completed futures."""
    idx, result_or_error = future.result()
    results[idx] = result_or_error

    if isinstance(result_or_error, RunErr):
        error_key = f"Error_{result_or_error.error_type}"
        result_counter[error_key] += 1
        if result_counter[error_key] == 1:
            logger.error(
                f"First error of type {result_or_error.error_type}: {result_or_error.message}"
            )
        if stop_on_error:
            stop_event.set()
    else:
        result_counter["SUCCESS"] += 1


def multi_thread(
    func: Callable,
    orig_inputs: List[Any],
    workers: int = 4,
    verbose: bool = None,
    desc: str = None,
    stop_on_error: bool = False,
    filter_none: bool = False,
    do_memoize: bool = False,
) -> List[Any]:
    """
    Execute the given function `func` on a list of inputs `orig_inputs` in parallel
    using multiple threads. Manages progress tracking, handling of errors, and optional
    filtering of None results.

    Parameters
    ----------
    func : Callable
        The function to be executed for each input.
    orig_inputs : List[Any]
        The list of inputs to process.
    workers : int, optional
        Number of worker threads to use. Defaults to 4. If set to 1, the execution is
        effectively single-threaded.
    verbose : bool, optional
        If True, shows a tqdm progress bar. If None, it defaults to True if `desc` is
        provided, otherwise False.
    desc : str, optional
        The description for the progress bar. Providing this automatically sets
        verbose to True if verbose is None.
    stop_on_error : bool, optional
        If True, stops execution upon the first exception encountered in any thread.
    filter_none : bool, optional
        If True, filters out None results from the final list.

    Returns
    -------
    List[Any]
        The list of results from applying `func` to each element in `orig_inputs`,
        in the original order. May exclude None values if `filter_none` is True.

    Notes
    -----
    - If the environment variable SPEEDY_DEBUG is set to "1", the function forces
      `workers` to 1 and runs in single-threaded mode, aiding debug.
    - Uses a ThreadPoolExecutor with as_completed to manage parallel tasks.
    - Displays progress via tqdm if verbose is True.
    """
    if do_memoize:
        # apply memoize on the function, need to make sure the method
        from speedy_utils import memoize

        func = memoize(func)
    if verbose is None:
        # Default verbosity based on whether a desc is provided
        verbose = bool(desc)

    if bool(int(os.getenv("SPEEDY_DEBUG", "0"))):
        logger.debug("Running in debug mode, setting workers to 1")
        workers = 1

    # If workers <= 1, no parallelization
    if workers <= 1:
        return [func(inp) for inp in tqdm(orig_inputs, desc="Single thread")]

    inputs = handle_inputs(func, orig_inputs)
    stop_event = threading.Event()
    results = [None] * len(inputs)
    result_counter = defaultdict(int)

    log_fn = logger.info if verbose else (lambda x: None)

    __execute_tasks_in_parallel(
        func,
        workers,
        verbose,
        desc,
        stop_on_error,
        inputs,
        stop_event,
        results,
        result_counter,
        log_fn,
    )

    if filter_none:
        results = [r for r in results if r is not None]

    return results


def __execute_tasks_in_parallel(
    func,
    workers,
    verbose,
    desc,
    stop_on_error,
    inputs,
    stop_event,
    results,
    result_counter,
    log_fn,
):
    try:
        with ThreadPoolExecutor(max_workers=workers) as executor:
            futures = [
                executor.submit(_process_single_task, func, i, inp, stop_event)
                for i, inp in enumerate(inputs)
            ]

            last_update_time = time.time()
            completed_since_last_update = 0

            with tqdm(
                total=len(futures), desc=desc, disable=not verbose, ncols=120
            ) as pbar:
                for future in as_completed(futures):
                    if stop_event.is_set():
                        logger.info("Stopping execution due to an error.")
                        break

                    _handle_results(
                        future, results, stop_event, stop_on_error, result_counter, pbar
                    )

                    completed_since_last_update += 1
                    current_time = time.time()

                    # Update progress bar and postfix periodically
                    if current_time - last_update_time > 0.1:
                        pbar.set_postfix(dict(result_counter))
                        pbar.update(completed_since_last_update)
                        completed_since_last_update = 0
                        last_update_time = current_time

                # Final update for any remaining tasks
                pbar.set_postfix(dict(result_counter))
                pbar.update(completed_since_last_update)

    except KeyboardInterrupt:
        logger.warning("Execution manually interrupted.")
        stop_event.set()

    except Exception as e:
        import traceback

        log_fn(f"An error occurred during execution: {traceback.format_exc()}")

    finally:
        if verbose:
            fprint(result_counter, "Result counter", is_notebook=False)



from collections import Counter
from typing import List, TypeVar, Generic, Callable, Any
from threading import Lock
from loguru import logger

T = TypeVar('T')

class TaskDistributor(Generic[T]):
    """Generic class for distributing tasks across multiple workers."""

    def __init__(self, workers: List[T], debug: bool = False):
        """
        Initialize with a list of workers and an optional debug flag.

        :param workers: List of worker instances.
        :param debug: If True, enables debug-level logging.
        """
        self.workers = workers
        self.usage_counter = Counter()
        for w in workers:
            self.usage_counter[w] = 0  # Initialize usage counter for each worker
        self.lock = Lock()  # Ensures thread-safe access to shared data

        # Set up logger with loguru
        self.debug = debug
        if self.debug:
            logger.enable(__name__)  # Enable logging for this module
            logger.debug("Logger initialized in debug mode.")
        else:
            logger.disable(__name__)  # Disable logging for this module when debug is False
    def _get_least_busy_worker(self) -> T:
        """Select the worker that has handled the fewest tasks."""
        with self.lock:
            v2k = {v: k for k, v in self.usage_counter.items()}
            min_value = min(v2k.keys())
            worker = v2k[min_value]
            logger.info(f"Selected worker: {worker}, usage_counter: {self.usage_counter}")
            return worker

    def _update_worker_usage(self, worker: T, delta: int):
        """Update the usage count of a worker."""
        with self.lock:
            self.usage_counter[worker] += delta
            action = "Incremented" if delta > 0 else "Decremented"
            logger.debug(f"{action} usage count for worker {worker}, Usage:{self.usage_counter}")

    def delegate_task(self, task_func: Callable[[T], Any], *args, **kwargs) -> Any:
        """
        Delegate a task to the least busy worker.

        :param task_func: The task function to execute.
        :param args: Positional arguments for the task function.
        :param kwargs: Keyword arguments for the task function.
        :return: The result of the task function.
        """
        worker = self._get_least_busy_worker()
        # self._update_worker_usage(worker, 1)  # Increment the usage counter
        logger.debug(f"Delegating task {task_func.__name__} to worker {worker} with args {args} and kwargs {kwargs}")
        try:
            result = task_func(worker, *args, **kwargs)
            logger.debug(f"Task {task_func.__name__} completed by worker {worker} with result {result}")
            return result
        except Exception as e:
            logger.error(f"Error executing task {task_func.__name__} with worker {worker}: {e}")
            raise
        # finally:
        #     self._update_worker_usage(worker, -1)  # Decrement the usage counter

    def __getattr__(self, name: str):
        """
        Dynamically delegates the method to the least busy worker.
        Called when an attribute is not found in TaskDistributor.

        :param name: The attribute name to delegate.
        :return: The delegated attribute or method.
        """
        try:
            worker = self._get_least_busy_worker()
            if hasattr(worker, name):
                attr = getattr(worker, name)
                # If it's a callable (method), return a wrapper to delegate the task
                if callable(attr):
                    def delegated_method(*args, **kwargs):
                        logger.debug(f"Delegating method '{name}' to worker {worker} with args {args} and kwargs {kwargs}")
                        return self.delegate_task(attr, *args, **kwargs)
                    return delegated_method
                else:

                    logger.debug(f"Accessing attribute '{name}' of worker {worker}")
                    return attr  # If it's not callable, return the attribute directly
            error_message = f"'{type(self).__name__}' object has no attribute '{name}'"
            logger.error(error_message)
            raise AttributeError(error_message)
        except:
            return super().__getattr__(name)

    def __dir__(self):
        """
        Include methods of the workers in TaskDistributor's `dir()`.
        This helps support autocompletion in IDEs.

        :return: Combined list of TaskDistributor and worker attributes.
        """
        worker_methods = dir(self._get_least_busy_worker())
        combined_dir = list(set(super().__dir__() + worker_methods))
        if self.debug:
            logger.debug(f"Combined dir includes: {combined_dir}")
        return combined_dir



from speedy_utils.common.dataclass_parser import ArgsParser
from speedy_utils.multi_worker.process import multi_process
from .common.clock import Clock, speedy_timer, timef
from .common.generic import TaskDistributor
from .common.utils_cache import (
    SPEED_CACHE_DIR,
    identify,
    identify_uuid,
    memoize,
)
from .common.utils_io import (
    dump_json_or_pickle,
    dump_jsonl,
    load_by_ext,
    load_json_or_pickle,
)
from .common.utils_misc import (
    convert_to_builtin_python,
    flatten_list,
    get_arg_names,
    is_notebook,
    mkdir_or_exist,
)
from .common.utils_print import fprint, print_table, setup_logger

from .multi_worker.thread import multi_thread

__all__ = [
    "SPEED_CACHE_DIR",
    "multi_process",
    "mkdir_or_exist",
    "dump_jsonl",
    "dump_json_or_pickle",
    "timef",  # Ensure timef is moved to an appropriate module or included here
    "load_json_or_pickle",
    "load_by_ext",
    "identify",
    "identify_uuid",
    "flatten_list",
    "fprint",
    "get_arg_names",
    "is_notebook",
    "print_table",
    "convert_to_builtin_python",
    "Clock",
    "multi_thread",
    "memoize",
    "speedy_timer",
    "TaskDistributor",
    "setup_logger",
    "ArgsParser",
]
__version__ = "0.1.0"



import functools
import inspect
from collections.abc import Iterable
from typing import Any, Callable, Dict, List, Union
import pandas as pd

# Example object


def _get_original_func(func):
    """
    Recursively unwrap a decorated function to find the actual
    original function object.
    """
    while hasattr(func, "__wrapped__"):
        func = func.__wrapped__
    return func


def handle_inputs(
    f: Callable, inputs: Union[List[Dict[str, Any]], List[Any], pd.DataFrame]
) -> List[Dict[str, Any]]:
    # 1. Unwrap in case f is decorated (e.g., by @memoize).
    real_func = _get_original_func(f)

    # 2. Count parameters with inspect.signature.
    #    This handles normal or annotated arguments, etc.
    sig = inspect.signature(real_func)
    num_params = len(sig.parameters)

    # Convert certain input types to list to unify processing
    if isinstance(inputs, (range, list, tuple)):
        inputs = list(inputs)

    # 3. If exactly 1 parameter, we do the single-arg logic:
    if num_params == 1:
        # If the user passed a dataframe, break it into rows
        if isinstance(inputs, pd.DataFrame):
            inputs = [r for _, r in inputs.iterrows()]

        # For a single-arg function, turn each item into a dict: {arg_name: item}
        # so we can later call func(**inp)
        arg_name = next(iter(sig.parameters))  # name of the single parameter
        inputs = [{arg_name: input_} for input_ in inputs]
        return inputs

    else:
        # If you really only plan to support single-arg calls, raise an error.
        # Or you could implement multi-arg logic here if desired.
        raise NotImplementedError(
            "Function has more than one argument, not implemented yet."
        )



# utils/utils_misc.py

import inspect
import os
import sys
from typing import Any, Callable, List
from IPython import get_ipython
from pydantic import BaseModel


def mkdir_or_exist(dir_name: str) -> None:
    """Create a directory if it doesn't exist."""
    os.makedirs(dir_name, exist_ok=True)


def flatten_list(list_of_lists: List[List[Any]]) -> List[Any]:
    """Flatten a list of lists into a single list."""
    return [item for sublist in list_of_lists for item in sublist]


def get_arg_names(func: Callable) -> List[str]:
    """Retrieve argument names of a function."""
    return inspect.getfullargspec(func).args


def is_notebook() -> bool:
    try:
        if "get_ipython" in globals().keys():
            get_ipython = globals()["get_ipython"]
            shell = get_ipython().__class__.__name__
            if shell == "ZMQInteractiveShell":
                return True  # Jupyter notebook or qtconsole
        return False  # Other type (?)
    except NameError:
        return False  # Probably standard Python interpreter


def convert_to_builtin_python(input_data: Any) -> Any:
    """Convert input data to built-in Python types."""
    if isinstance(input_data, dict):
        return {k: convert_to_builtin_python(v) for k, v in input_data.items()}
    elif isinstance(input_data, list):
        return [convert_to_builtin_python(v) for v in input_data]
    elif isinstance(input_data, (int, float, str, bool, type(None))):
        return input_data
    elif isinstance(input_data, BaseModel):
        data = input_data.model_dump_json()
        return convert_to_builtin_python(data)
    else:
        raise ValueError(f"Unsupported type {type(input_data)}")




import argparse
import yaml
from dataclasses import dataclass, fields, is_dataclass
from typing import Type, TypeVar, Any, Dict
from tabulate import tabulate

T = TypeVar("T")

class ArgsParser:
    @classmethod
    def get_parser(cls) -> argparse.ArgumentParser:
        """Generate an argument parser from the dataclass fields."""
        parser = argparse.ArgumentParser(description=f"Parser for {cls.__name__}")
        parser.add_argument("--yaml_file", type=str, help="Path to YAML file with arguments")

        for field in fields(cls):
            arg_name = f"--{field.name}"
            default = field.default
            field_type = field.type
            # print(field, field_type)
            if field_type is bool:
                parser.add_argument(arg_name, action="store_true", help=f"Enable {field.name} (default: {default})")
            elif 'list' in str(field_type):
                elem_type = str(field_type).split('[')[1].split(']')[0]
                parser.add_argument(arg_name, type=eval(elem_type), nargs='+', help=f"Override {field.name} (default: {default})")
            else:
                parser.add_argument(arg_name, type=field_type, default=None, help=f"Override {field.name} (default: {default})")
        return parser

    @classmethod
    def from_args(cls, args: argparse.Namespace) -> T:
        """Create an instance of the dataclass from argparse arguments."""
        config: Dict[str, Any] = {}
        if args.yaml_file:
            with open(args.yaml_file, "r") as file:
                config = yaml.safe_load(file)

        cli_overrides = {field.name: getattr(args, field.name) for field in fields(cls) if getattr(args, field.name) is not None}
        config.update(cli_overrides)

        return cls(**config)

    @classmethod
    def parse_args(cls) -> T:
        """Parse arguments and return an instance of the dataclass."""
        parser = cls.get_parser()
        args = parser.parse_args()
        return cls.from_args(args)

    def __str__(self):
        return tabulate([[f.name, getattr(self, f.name)] for f in fields(self)], headers=["Field", "Value"], tablefmt="github")

@dataclass
class ExampleArgs(ArgsParser):
    from_peft: str = "./outputs/llm_hn_qw32b/hn_results_r3/"
    model_name_or_path: str = "Qwen/Qwen2.5-32B-Instruct-AWQ"
    use_fp16: bool = False
    batch_size: int = 1
    max_length: int = 512
    cache_dir: str = ".cache/run_embeds"
    output_dir: str = ".cache"
    input_file: str = ".cache/doc.csv"
    output_name: str = "qw32b_r3"

if __name__ == "__main__":
    args = ExampleArgs.parse_args()
    print(args)



import gc
import multiprocessing
import time
from multiprocessing import Pool
from typing import Any, Callable, Dict, List

from loguru import logger
from tqdm import tqdm
import dill  # Add this import

from speedy_utils.multi_worker._handle_inputs import handle_inputs


def task_wrapper(func: Callable, index: int, input_kwargs: Dict[str, Any]) -> tuple:
    """
    Wraps the function execution to capture the result along with its index.
    """
    try:
        result = func(**input_kwargs)
        return index, result
    except Exception as e:
        logger.error(f"Error in task {index}: {e}")
        return index, None  # Handle exception and return None for failed tasks


def multi_process(
    func: Callable, inputs: List[Dict[str, Any]], workers: int = 4, verbose: bool = True
) -> List[Any]:
    """
    Executes a function concurrently across multiple processes with a list of dictionary inputs.
    Ensures ordered output by preserving task indices.
    """
    inputs = handle_inputs(func, inputs)

    with multiprocessing.Manager() as manager:
        # Create a pool of worker processes
        with Pool(processes=workers, initializer=dill.load, initargs=(dill.dumps(func),)) as pool:
            tasks = []
            try:
                # Prepare and submit tasks to the pool
                for idx, inp in enumerate(inputs):
                    task = pool.apply_async(task_wrapper, args=(func, idx, inp))
                    tasks.append(task)

                # Display a progress bar if verbose is True
                results = []
                with tqdm(total=len(tasks), desc="Processing", disable=not verbose) as pbar:
                    for task in tasks:
                        idx, result = task.get()  # Wait for each task to complete and retrieve its result
                        results.append((idx, result))
                        pbar.update(1)  # Update progress bar by 1 for each completed task

            except KeyboardInterrupt:
                logger.warning("KeyboardInterrupt detected, returning partial results...")
                pool.terminate()  # Stop remaining processes
            finally:
                logger.debug("Closing and joining the pool...")
                pool.close()  # Ensure all processes are properly closed
                pool.join()  # Wait for all processes to complete
                gc.collect()  # Collect garbage to free up resources

        # Sort results by task index to ensure ordered output
        results.sort(key=lambda x: x[0])
        ordered_results = [result for _, result in results]

        # Check if more than 5 percent of tasks failed, then raise a warning
        none_rate = sum(1 for r in ordered_results if r is None) / len(ordered_results)
        if none_rate > 0.05:
            logger.warning(
                f"{none_rate*100:0.2f} % of tasks failed. Consider increasing workers or checking input data."
            )

    return ordered_results


# Example usage
if __name__ == "__main__":
    class Test:
        def f_simple(self, **kwargs):
            """
            A simple function to simulate a task.
            """
            y = kwargs["y"]
            x = kwargs["x"]
            logger.info(f"Starting task with x={x}, y={y}, sleep {x}")
            time.sleep(x)  # Simulate a delay
            result = x / y
            logger.info(f"Done with x={x}, y={y}, result={result}")
            return result

    o = Test()
    inputs = [
        {"x": 1, "y": 2},
        {"x": 4, "y": 10},
        {"x": 3, "y": 1},
        {"x": 1, "y": 100},
        {"x": 5, "y": 6},
        {"x": 3, "y": 12},
    ]

    def f1(**kwargs):
        return o.f_simple(**kwargs)

    results = multi_process(f1, inputs, workers=3, verbose=True)
    print(results)


# utils/utils_print.py

import copy
import json
import pprint
import textwrap
from typing import Any, Dict, List, Literal, Optional

from IPython.display import HTML, display
from tabulate import tabulate
from loguru import logger


import sys
from loguru import logger

from .utils_misc import is_notebook


def display_pretty_table_html(data: Dict) -> None:
    """
    Display a pretty HTML table in Jupyter notebooks.
    """
    table = "<table>"
    for key, value in data.items():
        table += f"<tr><td>{key}</td><td>{value}</td></tr>"
    table += "</table>"
    display(HTML(table))


# Flattening the dictionary using "." notation for keys
def flatten_dict(d, parent_key="", sep="."):
    items = []
    for k, v in d.items():
        new_key = parent_key + sep + k if parent_key else k
        if isinstance(v, dict):
            items.extend(flatten_dict(v, new_key, sep=sep).items())
        else:
            items.append((new_key, v))
    return dict(items)


def fprint(
    input_data: Any,
    key_ignore: Optional[List[str]] = None,
    key_keep: Optional[List[str]] = None,
    max_width: int = 100,
    indent: int = 2,
    depth: Optional[int] = None,
    table_format: str = "grid",
    str_wrap_width: int = 80,
    grep=None,
    is_notebook=None,
    f=print,
) -> None | str:
    """
    Pretty print structured data.
    """
    from speedy_utils import is_notebook as is_interactive

    # is_notebook = is_notebook or is_interactive()
    if is_notebook is None:
        is_notebook = is_interactive()
    if hasattr(input_data, "toDict"):
        input_data = input_data.toDict()
    if hasattr(input_data, "to_dict"):
        input_data = input_data.to_dict()

    if hasattr(input_data, "model_dump"):
        input_data = input_data.model_dump()
    if not isinstance(input_data, (dict, str)):
        raise ValueError("Input data must be a dictionary or string")

    if isinstance(input_data, dict):
        input_data = flatten_dict(input_data)

    if grep is not None:
        input_data = {k: v for k, v in input_data.items() if grep in str(k)}

    def remove_keys(d: Dict, keys: List[str]) -> Dict:
        """Remove specified keys from a dictionary."""
        for key in keys:
            parts = key.split(".")
            sub_dict = d
            for part in parts[:-1]:
                sub_dict = sub_dict.get(part, {})
            sub_dict.pop(parts[-1], None)
        return d

    def keep_keys(d: Dict, keys: List[str]) -> Dict:
        """Keep only specified keys in a dictionary."""
        result = {}
        for key in keys:
            parts = key.split(".")
            sub_source = d
            sub_result = result
            for part in parts[:-1]:
                if part not in sub_source:
                    break
                sub_result = sub_result.setdefault(part, {})
                sub_source = sub_source[part]
            else:
                sub_result[parts[-1]] = copy.deepcopy(sub_source.get(parts[-1]))
        return result

    if hasattr(input_data, "to_dict"):
        input_data = input_data.to_dict()

    processed_data = copy.deepcopy(input_data)

    if isinstance(processed_data, dict) and is_notebook:
        if key_keep is not None:
            processed_data = keep_keys(processed_data, key_keep)
        elif key_ignore is not None:
            processed_data = remove_keys(processed_data, key_ignore)

        if is_notebook:
            display_pretty_table_html(processed_data)
            return

    if isinstance(processed_data, dict):
        table = [[k, v] for k, v in processed_data.items()]
        f(
            tabulate(
                table,
                headers=["Key", "Value"],
                tablefmt=table_format,
                maxcolwidths=[None, max_width],
            )
        )
    elif isinstance(processed_data, str):
        wrapped_text = textwrap.fill(processed_data, width=str_wrap_width)
        f(wrapped_text)
    elif isinstance(processed_data, list):
        f(tabulate(processed_data, tablefmt=table_format))
    else:
        printer = pprint.PrettyPrinter(width=max_width, indent=indent, depth=depth)
        printer.pprint(processed_data)


def print_table(data: Any) -> None:
    """
    Print data as a table.
    """

    def __get_table(data: Any) -> str:
        if isinstance(data, str):
            try:
                data = json.loads(data)
            except json.JSONDecodeError as exc:
                raise ValueError("String input could not be decoded as JSON") from exc

        if isinstance(data, list):
            if all(isinstance(item, dict) for item in data):
                headers = list(data[0].keys())
                rows = [list(item.values()) for item in data]
                return tabulate(rows, headers=headers)
            else:
                raise ValueError("List must contain dictionaries")

        if isinstance(data, dict):
            headers = ["Key", "Value"]
            rows = list(data.items())
            return tabulate(rows, headers=headers)

        raise TypeError(
            "Input data must be a list of dictionaries, a dictionary, or a JSON string"
        )

    table = __get_table(data)
    print(table)


from loguru import logger
import sys
import re

from loguru import logger
import sys
import re


def setup_logger(level: str = "INFO", enable_grep: str = "", disable_grep: str = ""):
    """
    Setup the logger with the specified level and control logging based on grep patterns.

    :param level: The desired log level.
                  Valid levels: 'T' (TRACE), 'D' (DEBUG), 'I' (INFO), 'S' (SUCCESS),
                  'W' (WARNING), 'E' (ERROR), 'C' (CRITICAL), 'DISABLE'.
    :param enable_grep: A comma-separated string of patterns. Only logs matching these patterns will be enabled.
    :param disable_grep: A comma-separated string of patterns. Logs matching these patterns will be disabled.
    """

    # Map the shorthand level to the full name
    level_mapping = {
        "T": "TRACE",
        "D": "DEBUG",
        "I": "INFO",
        "S": "SUCCESS",
        "W": "WARNING",
        "E": "ERROR",
        "C": "CRITICAL",
    }

    # Set the level based on the input (or default to INFO)
    level = level_mapping.get(level.upper(), level.upper())

    # Remove any existing handlers to avoid duplication
    logger.remove()

    # Grep pattern handling
    enable_patterns = [
        pattern.strip() for pattern in enable_grep.split(",") if pattern.strip()
    ]
    disable_patterns = [
        pattern.strip() for pattern in disable_grep.split(",") if pattern.strip()
    ]

    def log_filter(record):
        """
        This filter applies the logging level and the grep pattern matching.
        Only logs with a level >= the specified level and that match the enable/disable patterns will be logged.
        """
        log_message = f"{record['file']}:{record['line']} ({record['function']})"

        # Check if the log should be enabled or disabled based on the grep patterns
        if enable_patterns and not any(
            re.search(pattern, log_message) for pattern in enable_patterns
        ):
            return False  # If enable_grep is provided, log only if it matches
        if disable_patterns and any(
            re.search(pattern, log_message) for pattern in disable_patterns
        ):
            return False  # If disable_grep matches, don't log

        # Return True if the log level is >= the set level
        return record["level"].no >= logger.level(level).no

    # Add the handler to stdout with the log format and the custom filter
    logger.add(
        sys.stdout,
        colorize=True,
        format=(
            "<level>{level: <8}</level> | <cyan>{file}:{line} ({function})</cyan> - "
            "<level>{message}</level>"
        ),
        filter=log_filter,
    )

    # Handle the "DISABLE" level: if set, disable all logging
    if level.upper() == "DISABLE":
        logger.disable("")  # Disable all logging
        logger.info("Logging disabled")
    else:
        logger.enable("")  # Ensure logging is enabled
        logger.debug(f"Logging set to {level}")



# speedy_utils/all.py

import copy
import functools
import gc
import inspect
import json
import multiprocessing
import os
import os.path as osp
import pickle
import pprint
import re
import sys
import textwrap
import threading
import time
import traceback
import uuid
from collections import Counter
from concurrent.futures import ThreadPoolExecutor, as_completed
from glob import glob
from multiprocessing import Pool
from pathlib import Path
from threading import Lock
from typing import Any, Callable, Dict, Generic, List, Literal, Optional, TypeVar, Union

import pandas as pd
import xxhash
from IPython import get_ipython
from IPython.display import HTML, display
from loguru import logger
from openai import BaseModel
from tabulate import tabulate
from tqdm import tqdm
import numpy as np

__all__ = [
    "copy",
    "functools",
    "gc",
    "inspect",
    "json",
    "multiprocessing",
    "os",
    "osp",
    "pickle",
    "pprint",
    "re",
    "sys",
    "textwrap",
    "threading",
    "time",
    "traceback",
    "uuid",
    "Counter",
    "ThreadPoolExecutor",
    "as_completed",
    "glob",
    "Pool",
    "Path",
    "Lock",
    "Any",
    "Callable",
    "Dict",
    "Generic",
    "List",
    "Literal",
    "Optional",
    "TypeVar",
    "Union",
    "pd",
    "xxhash",
    "get_ipython",
    "HTML",
    "display",
    "logger",
    "BaseModel",
    "tabulate",
    "tqdm",
    "np",
]
from speedy_utils import *  # noqa: F403
from speedy_utils import __all__ as all_speedy_utils

__all__ += all_speedy_utils



import inspect
import os
import time

from loguru import logger
from tabulate import tabulate

__all__ = ["Clock", "timef"]


def timef(func):
    "Decorator to print the execution time of a function"

    def wrapper(*args, **kwargs):
        start_time = time.time()
        result = func(*args, **kwargs)
        end_time = time.time()
        execution_time = end_time - start_time
        logger.opt(depth=2).info(
            f"{func.__name__} took {execution_time:0.2f} seconds to execute."
        )
        return result

    return wrapper


class Clock:
    """
    A simple timer utility to measure and log time intervals.

    Usage:

    1. Creating and starting the timer:
        timer = Timer(start_now=True)
        # or
        timer = Timer(start_now=False)
        timer.start()

    2. Measure time since the timer started:
        elapsed_time = timer.elapsed_time()

    3. Log the time elapsed since the timer started:
        timer.log_elapsed_time()
        # or use a custom logger
        timer.log_elapsed_time(custom_logger=my_custom_logger)

    4. Measure time since the last checkpoint:
        time_since_last_checkpoint = timer.time_since_last_checkpoint()

    5. Update a named task in the internal task time table:
        timer.update_task("task_name")

    6. Print the task time table every 'interval' seconds:
        timer.print_task_table(interval=1)
    """

    def __init__(self, start_now=True):
        """Initialize the timer and optionally start it immediately."""
        self.start_time = None
        self.task_times = {}
        self.last_checkpoint = None
        if start_now:
            self.start()
        self.print_counter = 0
        self.last_print_time = time.time()
        self.min_depth = float("inf")

    def start(self):
        """Start the timer or reset if already started."""
        if self.start_time is not None:
            raise ValueError("Timer has already been started.")
        self.start_time = time.time()
        self.last_checkpoint = self.start_time
        # logger.opt(depth=2).info(f"Timer started. {id(self)=}")

    def elapsed_time(self):
        """Return the time elapsed since the timer started."""
        if self.start_time is None:
            raise ValueError("Timer has not been started.")
        return time.time() - self.start_time

    def log_elapsed_time(self, custom_logger=None):
        """Log the time elapsed since the timer started."""
        msg = f"Time elapsed: {self.elapsed_time():.2f} seconds."
        if custom_logger:
            custom_logger(msg)
        else:
            logger.opt(depth=2).info(msg)

    def _tick(self):
        """Return the time elapsed since the last checkpoint and update the last checkpoint."""
        # assert self.start_time is not None, f"Timer has not been started. {id(self)=}"
        if not self.start_time:
            logger.opt(depth=2).warning(
                "Timer has not been started. Please call start() before using this method."
            )
            return
        current_time = time.time()
        elapsed = current_time - self.last_checkpoint
        self.last_checkpoint = current_time
        return elapsed

    def tick(self):
        return self._tick()
    
    def time_since_last_checkpoint(self):
        """Return the time elapsed since the last checkpoint."""
        if self.start_time is None:
            # raise ValueError("Timer has not been started.")
            logger.opt(depth=2).warning(
                "Timer has not been started. Please call start() before using this method."
            )
            return
        return time.time() - self.last_checkpoint

    def update_task(self, task_name):
        """Update the elapsed time for the specified task, including file, line, and call depth."""

        # Get the full call stack
        stack = inspect.stack()

        # Get the file and line number of the caller (the previous frame in the stack)
        caller_frame = stack[1]
        file_lineno = f"{os.path.basename(caller_frame.filename)}:{caller_frame.lineno}"

        # Calculate the depth of the current call (i.e., how far it is in the stack)
        call_depth = (
            len(stack) - 1
        )  # Subtract 1 to exclude the current frame from the depth count
        if call_depth < self.min_depth:
            self.min_depth = call_depth

        # Update the task time in the internal task table
        if task_name not in self.task_times:
            self.task_times[task_name] = {
                "time": 0,
                "file_lineno": file_lineno,
                "depth": call_depth,
            }
        self.task_times[task_name]["time"] += self.tick()

    def get_percentage_color(self, percentage):
        """Return ANSI color code based on percentage."""
        if percentage >= 75:
            return "\033[91m"  # Red
        elif percentage >= 50:
            return "\033[93m"  # Yellow
        elif percentage >= 25:
            return "\033[92m"  # Green
        else:
            return "\033[94m"  # Blue

    def print_task_table(self, interval=1, max_depth=None):
        """Print the task time table at regular intervals."""
        current_time = time.time()

        if current_time - self.last_print_time > interval:
            self.print_counter += 1
            total_time = (
                sum(data["time"] for data in self.task_times.values()) or 1
            )  # Avoid division by zero

            # Prepare data for the table
            table_data = []
            for task_name, data in self.task_times.items():
                time_spent = data["time"]
                file_lineno = data["file_lineno"]
                depth = data["depth"] - self.min_depth
                if max_depth is not None and depth > max_depth:
                    continue
                percentage = (time_spent / total_time) * 100

                # Get color code based on percentage
                color_code = self.get_percentage_color(percentage)
                percentage_str = f"{percentage:.2f} %"
                colored_percentage = f"{color_code}{percentage_str}\033[0m"

                table_data.append(
                    [
                        task_name,
                        file_lineno,
                        depth,
                        f"{time_spent:.2f} s",
                        colored_percentage,
                    ]
                )

            # Add headers and log using tabulate
            table = tabulate(
                table_data,
                headers=["Task", "File:Line", "Depth", "Time (s)", "Percentage (%)"],
                tablefmt="grid",
            )

            self.last_print_time = current_time
            total_time_str = f"\nTotal time elapsed: {total_time:.2f} seconds."
            logger.opt(depth=2).info(f"\n{table}\n{total_time_str}")


# Example of how to instantiate the Timer
speedy_timer = Clock(start_now=False)



# utils/utils_io.py

import json
import os
import os.path as osp
from pathlib import Path
import pickle
from glob import glob
from typing import Any, List, Dict, Union
import time
from .utils_misc import mkdir_or_exist


def dump_jsonl(list_dictionaries: List[Dict], file_name: str = "output.jsonl") -> None:
    """
    Dumps a list of dictionaries to a file in JSON Lines format.
    """
    with open(file_name, "w", encoding="utf-8") as file:
        for dictionary in list_dictionaries:
            file.write(json.dumps(dictionary, ensure_ascii=False) + "\n")


def dump_json_or_pickle(
    obj: Any, fname: str, ensure_ascii: bool = False, indent: int = 4
) -> None:
    """
    Dump an object to a file, supporting both JSON and pickle formats.
    """
    if isinstance(fname, Path):
        fname = str(fname)
    mkdir_or_exist(osp.abspath(os.path.dirname(osp.abspath(fname))))
    if fname.endswith(".json"):
        with open(fname, "w", encoding="utf-8") as f:
            try:
                json.dump(obj, f, ensure_ascii=ensure_ascii, indent=indent)
            # TypeError: Object of type datetime is not JSON serializable
            except TypeError:
                print(
                    "Error: Object of type datetime is not JSON serializable",
                    str(obj)[:1000],
                )
                raise
    elif fname.endswith(".jsonl"):
        dump_jsonl(obj, fname)
    elif fname.endswith(".pkl"):
        with open(fname, "wb") as f:
            pickle.dump(obj, f)
    else:
        raise NotImplementedError(f"File type {fname} not supported")


def load_json_or_pickle(fname: str, counter=0) -> Any:
    """
    Load an object from a file, supporting both JSON and pickle formats.
    """
    if fname.endswith(".json") or fname.endswith(".jsonl"):
        with open(fname, "r", encoding="utf-8") as f:
            return json.load(f)
    else:
        try:
            with open(fname, "rb") as f:
                return pickle.load(f)
        # EOFError: Ran out of input
        except EOFError:
            time.sleep(1)
            if counter > 5:
                print("Error: Ran out of input", fname)
                os.remove(fname)
                raise
            return load_json_or_pickle(fname, counter + 1)
        except Exception as e:
            raise ValueError(f"Error {e} while loading {fname}") from e


def load_jsonl(path):
    lines = open(path, "r", encoding="utf-8").read().splitlines()
    return [json.loads(line) for line in lines]


def load_by_ext(fname: Union[str, List[str]], do_memoize: bool = False) -> Any:
    """
    Load data based on file extension.
    """
    if isinstance(fname, Path):
        fname = str(fname)
    from .utils_cache import (
        memoize,
    )  # Adjust import based on your actual multi_worker module


    try:
        if isinstance(fname, str) and "*" in fname:
            paths = glob(fname)
            paths = sorted(paths)
            return multi_process(load_by_ext, paths, workers=16)
        elif isinstance(fname, list):
            paths = fname
            return multi_process(load_by_ext, paths, workers=16)

        def load_csv(path: str, **pd_kwargs) -> Any:
            import pandas as pd

            return pd.read_csv(path, engine="pyarrow", **pd_kwargs)

        def load_txt(path: str) -> List[str]:
            with open(path, "r", encoding="utf-8") as f:
                return f.read().splitlines()

        def load_default(path: str) -> Any:
            if path.endswith(".jsonl"):
                return load_jsonl(path)
            elif path.endswith(".json"):
                try:
                    return load_json_or_pickle(path)
                except json.JSONDecodeError as exc:
                    raise ValueError("JSON decoding failed") from exc
            return load_json_or_pickle(path)

        handlers = {
            ".csv": load_csv,
            ".tsv": load_csv,
            ".txt": load_txt,
            ".pkl": load_default,
            ".json": load_default,
            ".jsonl": load_default,
        }

        ext = os.path.splitext(fname)[-1]
        load_fn = handlers.get(ext)

        if not load_fn:
            raise NotImplementedError(f"File type {ext} not supported")

        if do_memoize:
            load_fn = memoize(load_fn)

        return load_fn(fname)
    except Exception as e:
        raise ValueError(f"Error {e} while loading {fname}") from e
